"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const chokidar_1 = require("chokidar");
const fm = require("front-matter");
const marked = require("marked");
const debug = require("debug");
const util = require("./util");
const get_image_meta_1 = require("./get-image-meta");
// TOOD: use ESModule syntax
const globby = require('globby');
const log = debug('phox:get-data');
const caches = {
    pageCache: {},
    imageCache: {},
    tagCache: [],
};
const getDirName = (pathName) => path
    .dirname(pathName)
    .split(path.sep)
    .pop();
const sortTagsByTitle = (a, b) => a.title.localeCompare(b.title);
const sortImagesByTitle = (a, b) => a.meta.title.localeCompare(b.meta.title);
const sortTagCacheAlphabetically = (t) => t.sort(sortTagsByTitle).map((tag) => (Object.assign({}, tag, { images: tag.images.sort(sortImagesByTitle) })));
const getIndexOfTagInCache = (slug, t) => t.findIndex((item) => item.slug === slug);
const getIndexOfImageInTagCache = (t, tagIndex, filePath) => t[tagIndex].images.findIndex((img) => img.filePath === filePath);
const removeFromCache = (allCaches, cacheType) => (filePath) => {
    if (!filePath.endsWith('.md') && !filePath.endsWith('.jpg')) {
        return;
    }
    const relativeFilePath = util.stripSlashes(filePath.replace(process.cwd(), ''));
    const cacheObject = allCaches[cacheType];
    const item = cacheObject[relativeFilePath];
    if (!Boolean(item)) {
        return;
    }
    log('Purge cache for file "%s".', filePath);
    allCaches[cacheType] = Object.keys(cacheObject)
        .filter((f) => f !== relativeFilePath)
        .reduce((acc, f) => (Object.assign({ [f]: cacheObject[f] }, acc)), {});
};
const addImageToTagCache = (image) => {
    log('Add image "%s" to cache.', image.filePath);
    return image.meta.tags.reduce((acc, tag) => {
        const tagIndex = getIndexOfTagInCache(tag.slug, acc);
        if (tagIndex === -1) {
            const tagCacheItem = Object.assign({}, tag, { images: [image] });
            return acc.concat(tagCacheItem);
        }
        const imageIndex = getIndexOfImageInTagCache(acc, tagIndex, image.filePath);
        if (imageIndex === -1) {
            acc[tagIndex].images = acc[tagIndex].images.concat(image);
        }
        else {
            acc[tagIndex].images = acc[tagIndex].images
                .slice(0, imageIndex)
                .concat([image].concat(acc[tagIndex].images.slice(imageIndex + 1)));
        }
        return acc;
    }, [...caches.tagCache]);
};
const removeImageFromTagCache = (filePath) => {
    if (!filePath.endsWith('.jpg')) {
        return;
    }
    const relativeFilePath = util.stripSlashes(filePath.replace(process.cwd(), ''));
    log('Remove image "%s" from cache.', filePath);
    caches.tagCache = sortTagCacheAlphabetically(caches.tagCache.reduce((acc, tag, i) => {
        const imageIndex = getIndexOfImageInTagCache(acc, i, relativeFilePath);
        if (imageIndex === -1) {
            return acc;
        }
        if (tag.images.length === 1) {
            acc[i].images = [];
            return acc;
        }
        acc[i].images = tag.images
            .slice(0, imageIndex)
            .concat(tag.images.slice(imageIndex + 1));
        return acc;
    }, [...caches.tagCache]));
};
const getDataForImage = (albumsDir, albumName) => (file) => ({
    filePath: file,
    detailLinkProps: util.getDetailLinkProps(albumsDir, albumName, file),
    fileName: path.basename(file),
});
const getMetaFromImage = (image) => __awaiter(this, void 0, void 0, function* () {
    const strippedImagePath = util.stripSlashes(image.filePath);
    const imageCache = caches.imageCache;
    if (imageCache[strippedImagePath]) {
        return imageCache[strippedImagePath];
    }
    const meta = yield get_image_meta_1.default(image.filePath);
    const imageWithMeta = Object.assign({}, image, { meta });
    caches.imageCache = Object.assign({}, imageCache, { [strippedImagePath]: imageWithMeta });
    caches.tagCache = sortTagCacheAlphabetically(addImageToTagCache(imageWithMeta));
    return imageWithMeta;
});
exports.getImages = (albumsDir, albumName) => __awaiter(this, void 0, void 0, function* () {
    const pattern = path.join('static', albumsDir, albumName, '*.jpg');
    const files = yield globby(pattern);
    const images = files.map(getDataForImage(albumsDir, albumName));
    log('Retrieved %d images for pattern "%s".', images.length, pattern);
    const imagesWithMeta = yield Promise.all(images.map(getMetaFromImage));
    return imagesWithMeta.sort((a, b) => a.fileName.localeCompare(b.fileName));
});
exports.getPageContent = (pagePath, pathPrefix = '') => __awaiter(this, void 0, void 0, function* () {
    log('Requesting page "%s"...', pagePath);
    const strippedPagePath = util.stripSlashes(pagePath);
    const pageCache = caches.pageCache;
    if (pageCache[strippedPagePath]) {
        log('Serve page from cache.', pagePath);
        return pageCache[strippedPagePath];
    }
    const content = yield util.pReadFile(pagePath);
    const { attributes, body } = fm(String(content));
    const name = getDirName(pagePath);
    caches.pageCache = Object.assign({}, pageCache, { [strippedPagePath]: {
            meta: attributes,
            path: `/${pathPrefix}${name}/`,
            body: marked(body),
            name: name === 'content' ? 'frontpage' : name,
        } });
    log('Freshly fetched page from the file system.');
    return caches.pageCache[strippedPagePath];
});
const getAlbums = (albumsDir) => (albumFile) => __awaiter(this, void 0, void 0, function* () {
    log('Requesting album "%s".', albumFile);
    const albumName = getDirName(albumFile);
    const [content, images] = yield Promise.all([
        exports.getPageContent(albumFile, `${albumsDir}/`),
        exports.getImages(albumsDir, albumName),
    ]);
    log('Found %d images for album "%s".', images.length, content.meta.title);
    return { content, images };
});
const getPages = (pagesGlob, albumsGlob) => __awaiter(this, void 0, void 0, function* () {
    const files = yield globby(pagesGlob, { ignore: [albumsGlob] });
    const pages = yield Promise.all(files.map((file) => __awaiter(this, void 0, void 0, function* () { return exports.getPageContent(file); })));
    return pages.sort((a, b) => a.name.localeCompare(b.name));
});
exports.getFrontpageApiData = (config) => __awaiter(this, void 0, void 0, function* () {
    const albumList = yield globby(util.getGlobPatterns(config).albums);
    const albumData = yield Promise.all(albumList.map((albumPath) => __awaiter(this, void 0, void 0, function* () { return exports.getPageContent(albumPath, `${config.albumsDir}/`); })));
    const albums = albumData
        .map(({ meta, name }) => ({
        linkProps: util.getAlbumLinkProps(config.albumsDir, name),
        meta: Object.assign({}, meta, { name }),
    }))
        .sort((a, b) => a.meta.title.localeCompare(b.meta.title));
    const content = yield exports.getPageContent(`${config.contentDir}/index.md`);
    return { albums, content };
});
exports.initCachePurger = (config) => {
    log('Initialize cache purger.');
    const { pages } = util.getGlobPatterns(config);
    const images = path.join('static', config.albumsDir, '**/*.jpg');
    const pagesMonitor = chokidar_1.watch(pages);
    const imagesMonitor = chokidar_1.watch(images);
    pagesMonitor.on('change', removeFromCache(caches, 'pageCache'));
    pagesMonitor.on('unlink', removeFromCache(caches, 'pageCache'));
    imagesMonitor.on('change', removeFromCache(caches, 'imageCache'));
    imagesMonitor.on('unlink', removeFromCache(caches, 'imageCache'));
    imagesMonitor.on('change', removeImageFromTagCache);
    imagesMonitor.on('unlink', removeImageFromTagCache);
};
exports.getDataforTag = (tagSlug) => __awaiter(this, void 0, void 0, function* () {
    const cache = caches.tagCache;
    const tagIndex = getIndexOfTagInCache(tagSlug, cache);
    if (tagIndex === -1) {
        log('No cache data available for tag slug "%s".', tagSlug);
        return;
    }
    log('Found tag "%s" in cache.', cache[tagIndex].title);
    return cache[tagIndex];
});
exports.default = (config) => __awaiter(this, void 0, void 0, function* () {
    log('Retrieving complete site data.');
    const patterns = util.getGlobPatterns(config);
    const albumFiles = yield globby(patterns.albums);
    const albums = yield Promise.all(albumFiles.map(getAlbums(config.albumsDir)));
    const pages = yield getPages(patterns.pages, patterns.albums);
    const tags = caches.tagCache;
    log('Albums: %O', albums);
    log('Pages: %O', pages);
    log('Tags: %O', tags);
    return { albums, pages, tags };
});
