"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const exif_1 = require("exif");
const marked = require("marked");
const slug = require("slug");
const debug = require("debug");
const nodeIptc = require("node-iptc");
const getImageSize = require("probe-image-size");
const utf8_1 = require("utf8");
const util_1 = require("./util");
const ORIENTATION_SQUARE = 'square';
const ORIENTATION_LANDSCAPE = 'landscape';
const ORIENTATION_PORTRAIT = 'portrait';
const NO_EXIF_SEGMENT = 'NO_EXIF_SEGMENT';
const log = debug('phox:getimage-meta');
const isError = (err) => err && err.code !== NO_EXIF_SEGMENT;
const hasNoExifData = (err) => err && err.code === NO_EXIF_SEGMENT;
const toTagData = (tag) => ({
    slug: slug(tag),
    title: tag,
});
const getExifData = (filePath) => __awaiter(this, void 0, void 0, function* () {
    return new Promise((resolve, reject) => {
        const args = { image: filePath };
        const extractor = new exif_1.ExifImage(args, (err, data) => {
            if (isError(err)) {
                reject({});
                return;
            }
            const exif = hasNoExifData(err) ? extractor.exifData : data;
            resolve(exif);
        });
    });
});
const getIptcData = (filePath) => __awaiter(this, void 0, void 0, function* () {
    try {
        const data = yield util_1.pReadFile(filePath);
        return nodeIptc(data) || {};
    }
    catch (e) {
        return {};
    }
});
const getDimensions = (filePath) => __awaiter(this, void 0, void 0, function* () {
    const input = fs_1.createReadStream(filePath);
    let orientation;
    let size;
    try {
        size = yield getImageSize(input);
        const { width, height } = size;
        if (width === height) {
            orientation = ORIENTATION_SQUARE;
        }
        if (width > height) {
            orientation = ORIENTATION_LANDSCAPE;
        }
        else {
            orientation = ORIENTATION_PORTRAIT;
        }
    }
    catch (e) { } // tslint:disable-line:no-empty
    input.destroy();
    return {
        width: size.width,
        height: size.height,
        orientation,
    };
});
const getCreationDateFromString = (date) => {
    const year = date.slice(0, 4);
    const month = date.slice(4, 6);
    const day = date.slice(6);
    return new Date(`${year}-${month}-${day}`)
        .toUTCString()
        .replace(/GMT.*$/, 'GMT');
};
const coordToDecimal = (gps) => {
    const latArr = gps.GPSLatitude;
    const lngArr = gps.GPSLongitude;
    if (!latArr || !lngArr) {
        return {};
    }
    const latRef = gps.GPSLatitudeRef || 'N';
    const lngRef = gps.GPSLongitudeRef || 'W';
    const lat = (latArr[0] + latArr[1] / 60 + latArr[2] / 3600) * (latRef === 'N' ? 1 : -1);
    const lng = (lngArr[0] + lngArr[1] / 60 + lngArr[2] / 3600) * (lngRef === 'W' ? -1 : 1);
    return { lat, lng };
};
const getDetailsFromMeta = (exif, iptc, dimensions) => (Object.assign({}, dimensions, { title: utf8_1.decode(iptc.object_name || ''), tags: (iptc.keywords || [])
        .map(utf8_1.decode)
        .filter(Boolean)
        .sort(util_1.sortAlphabetically)
        .map(toTagData), description: marked(utf8_1.decode(iptc.caption || '')), createdAt: iptc.date_created
        ? getCreationDateFromString(iptc.date_created)
        : undefined, camera: exif.image.Model || '', lens: exif.exif.LensModel || '', iso: exif.exif.ISO ? Number(exif.exif.ISO) : undefined, aperture: exif.exif.FNumber ? exif.exif.FNumber.toFixed(1) : undefined, focalLength: exif.exif.FocalLength
        ? exif.exif.FocalLength.toFixed(1)
        : undefined, exposureTime: exif.exif.ExposureTime
        ? Number(exif.exif.ExposureTime)
        : undefined, flash: Boolean(exif.exif.Flash), gps: coordToDecimal(exif.gps) }));
exports.default = (filePath) => __awaiter(this, void 0, void 0, function* () {
    log('Get image met for file "%s".', filePath);
    const [exif, iptc, dimensions] = yield Promise.all([
        getExifData(filePath),
        getIptcData(filePath),
        getDimensions(filePath),
    ]);
    return getDetailsFromMeta(exif, iptc, dimensions);
});
